"""Various tools for inverting terrain properties from signal components
Author: Cyril Grima <cyril.grima@gmail.com>
"""

import numpy as np
import subradar as sr
from scipy import constants as ct


def spm(frq, pc, pn):
    """Use the Small Perturbation Model with large correlation length assumption

    Arguments
    ---------
    frq : float
        radar frequency [Hz]
    pc : float
        calibrated coherent component [dB]
    pn : float
        calibrated incoherent component [dB]
    """

    wl = ct.c/frq
    k= 2*ct.pi/wl

    pc_lin = 10**(pc/10.)
    pn_lin = 10**(pn/10.)
    u = pc_lin/pn_lin

    def u_equation(k, sh):
        a = 2*k*sh
        return np.exp(-a**2)/a**2

    sh = wl*.001
    while u_equation(k, sh) > u:
        sh = sh+wl*.001

    r = -np.sqrt( pc_lin*np.exp((2*k*sh)**2) )
    eps = (1-r)**2/(1+r)**2

    return {'eps':eps, 'sh':sh}


def srf_coeff(Psc=None, Psn=None, h0=None, wb=None):
    """Invert the received powers (coherent, incoherent) into surface coefficients
    """
    L = sr.utils.geo_loss

    Psc = 10**(Psc/10.)
    Psn = 10**(Psn/10.)

    # Footprints
    As = 2*np.pi*sr.utils.footprint_rad_pulse(h0, wb)**2

    # Surface coefficients
    Rsc = Psc/L(2*h0)
    Rsn = Psn/L(h0)**2/As

    return 10*np.log10(Rsc), 10*np.log10(Rsn)


def bed_coeff(Psc=None, Psn=None, Pbc=None, Pbn=None, n1=None, sh=None,
              h0=None, h1=None, Q1=None, wf=None, wb=None):
    """Use the surface properties and bed powers to get bed reflectivity and
    backscatter coefficients. Signal is assumed to be already calibrated.
    Arguments must be energy in dB
    """
    L = sr.utils.geo_loss

    wk = sr.utils.wf2wk(wf)

    Rsc, Rsn = srf_coeff(Psc=Psc, Psn=Psn, h0=h0, wb=wb)

    Psc = 10**(Psc/10.)
    Psn = 10**(Psn/10.)
    Pbc = 10**(Pbc/10.)
    Pbn = 10**(Pbn/10.)
    Rsc = 10**(Rsc/10.)
    Rsn = 10**(Rsn/10.)
    Q1 = 10**(Q1/10.)

    # Footprints
    As = 2*np.pi*sr.utils.footprint_rad_pulse(h0, wb)**2
    Ab = 2*np.pi*sr.utils.footprint_rad_pulse(h0+h1/n1, wb)**2

    Tsc = np.abs(4*n1/(1+n1)**2) * np.exp(-wk**2*sh**2*(1-n1)**2)
    Tsn = n1*Rsn

    #Coefficients
    b1 = L(2*h0+2*h1)            *Tsc**2

    Rbc = Pbc/b1/Q1**2

    b2 = L(h0+2*h1)*L(h0)        *Tsc*Tsn*As
    b3 = L(h0+h1)*L(h0+h1)       *Tsc**2*Ab
    b4 = L(h0+h1)*L(h0)*L(h1)    *Tsc*Tsn*As*Ab
    b5 = b2
    b6 = L(h0)*L(2*h1)*L(h0)     *Tsn**2*As**2
    b7 = b4
    b8 = L(h0)*L(h1)*L(h0)*L(h1) *Tsn**2*As**2*Ab

    Rbn = (Pbn/Q1**2-Rbc*(b2+b5+b6)) / (b3+b4+b7+b8)

    return 10*np.log10(Rbc), 10*np.log10(Rbn)



"""
Probability density functions compliant with the lmfit package
Author: Cyril Grima <cyril.grima@gmail.com>
"""

import math
import mpmath
import numpy as np
from scipy import stats, integrate
from scipy.special import jv, kv, j0, i0, digamma, factorial, kn



def gamma(params, x, data=None, eps=None):
    """Gamma PDF from scipy with adequate variables
    """
    # Initialisation
    mu = params['mu']
    # Debug inputs
    if hasattr(mu, 'value'):
        mu = mu.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element
    # Model function
    model = stats.gamma.pdf(x, mu, scale = 1)
    model = np.nan_to_num(model, posinf=np.nan)

    if data is None:
        return model
    if eps is None:
        return model - data #residual
    return (model - data)/eps



def rayleigh(params, x, data=None, eps=None):
    """Rayleigh PDF from scipy with adequate variables
    """
    # Initialisation
    s = params['s']
    # Debug inputs
    if hasattr(s, 'value'): s = s.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element
    # Model function
    model = stats.rayleigh.pdf(x, scale = s)
    model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return model - data #residual
    return (model - data)/eps



def rice(params, x, data=None, eps=None, method='scipy'):
    """Rice PDF from scipy with adequate variables
    """
    # Initialisation
    a = params['a']
    s = params['s']
    # Debug inputs
    if hasattr(a, 'value'): a = a.value #debug due to lmfit.minimize
    if hasattr(s, 'value'): s = s.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element
    # Model function
    if method == 'scipy':
        model = stats.rice.pdf(x, a/s, scale = s)
    if method == 'analytic':
        model = (x/s**2) * np.exp(-(x**2+a**2)/(2*s**2)) * np.i0(a*x/s**2)
    #model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return model - data #residual
    return (model - data)/eps



def k(params, x, data=None, eps=None):
    """K PDF
    """
    # Initialisation
    s = params['s']
    mu = params['mu']
    # Debug inputs
    if hasattr(s, 'value'): s = s.value #idem
    if hasattr(mu, 'value'): mu = mu.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element

    # Model function
    b = np.sqrt(2*mu)/s
    model = 2*(x/2.)**mu *b**(mu+1.) /math.gamma(mu) *kv(mu-1,b*x)
    model = np.nan_to_num(model)

    if data is None:
        return model
    if eps is None:
        return model - data #residual
    return (model - data)/eps


def hk_auto(params, x, **kwargs):
    """ !!! EXPERIMENTAL !!
    Selector and application of the best method for HK.
    the 'analytic' method is privileged over 'compound' when it is
    stable enough because it is faster. Valid when :
         (Pc + Pn) <= 1
            AND
    -40 dB < Pc/Pn > 25 dB
    """
    pt = 10*np.log10(params['a']**2 + 2*params['s']**2*params['mu'])
    if (params['mu'] > 1) and (pt > -20):
        kwargs['method'] = 'analytic'
    else:
        kwargs['method'] = 'compound'

    return hk(params, x, **kwargs)


def hk(params, x, data=None, eps=None, method = 'analytic'):
    """Homodyne K-distribution from various methods

    Arguments
    ---------
    params : dict
        params for the hk function {'a', 's', 'mu'}
    x : sequence
        x coordinates

    Keywords
    --------
    data : sequence
        data to compare the result with (for minimization)
    eps : sequence
        error on the data
    method : string
        'auto' = Choose the best method from the range of input parameters.
        'analytic' = from the common analytic form Destrempes and Cloutier [2010].
                     Instable for mu < 1.
        'compound' = from the compound representation [Destrempes and Cloutier,
                     2010, Ultrasound in Med. and Biol. 36(7): 1037-51, Eq. 16].
                     Gives best results , especially for mu < 1 but is ~10x longer!
        'drumheller' = from Drumheller [2002, Eq. 34].
                       For testing only, not advised.
    verbose : bool
        print fitting report
    """
    if method == 'auto':
        return hk_auto(params, x, data=None, eps=None)

    # Initialisation
    a = params['a']
    s = params['s']
    mu = params['mu']

    # Debug inputs
    if hasattr(a, 'value'):
        a = a.value #debug due to lmfit.minimize
    if hasattr(s, 'value'):
        s = s.value #idem
    if hasattr(mu, 'value'):
        mu = mu.value #idem
    x = np.array([x]).flatten('C') # debug for iteration over 0-d element

    def integrand_analytic(w, x, a, s, mu):
        return x*w *j0(w*a) *j0(w*x) *(1. +w**2*s**2/2.)**-mu

    def integrand_compound(w, x, a, s, mu):
        r = rice({'a':a,'s':np.sqrt(s**2*w)}, x, method='scipy')
        g = gamma({'mu':mu}, w)
        return r*g

    def integrand_drumheller(w, x, a, s, mu):
        w = np.float64(w) #Convert mpmath to float64 type. Solve 'u_fonc not supported' error
        # Parameters correspondance
        r = x
        r0 = a
        v = mu-1
        b2 = 2/s**2
        b = np.sqrt(b2)
        # Function
        f1 = b2*r/math.gamma(v+1)
        f2 = (b2*r0*r)**(2*w) / 4**(2*w) / factorial(w)**2
        f3 = (b*np.sqrt(r0**2+r**2)/2.)**(v-2*w)
        f4 = kn(2*w-v, b*np.sqrt(r0**2+r**2))
        return f1*f2*f3*f4

    integrands = {
        'analytic': integrand_analytic,
        'compound': integrand_compound,
        'drumheller':integrand_drumheller,
    }
    integrand = integrands[method]
    # W0640: Cell variable i defined in loop (cell-var-from-loop)
    if method == 'drumheller':
        model = [mpmath.nsum(lambda w: integrand_drumheller(w, i, a, s, mu), [0, np.inf])
                for i in x] # Summation
    else:
        model = [integrate.quad(integrand, 0., np.inf, args=(i, a, s, mu), full_output=1)[0]
                 for i in x] # Integration

    model = np.array(model)

    # Set as nan where gamma distribution is not defined
    g = gamma({'mu':mu}, x)
    model[np.isfinite(g) == False] = np.nan

    if data is None:
        return model
    if eps is None:
        return model - data #residual
    return (model - data)/eps



"""
Various tools for extracting signal components from a fit of the amplitude
distribution
"""

import time
#import random

import numpy as np
import matplotlib
matplotlib.use('Agg')
#import matplotlib.pyplot as plt
from lmfit import minimize, Parameters, report_fit

from . import pdf
from .Classdef import Statfit


def param0(sample, method='basic'):
    """Estimate initial parameters for HK fitting

    Arguments
    ---------
    sample : sequence
        amplitudes

    Keywords
    --------
    method : string
        method to compute the initial parameters
    """
    if method == 'basic':
        a = np.nanmean(sample)
        s = np.nanstd(sample)
        mu = 1.
    return {'a':a, 's':s, 'mu':mu}


def lmfit(sample, fit_model='hk', bins='auto', p0 = None,
          xtol=1e-4, ftol=1e-4):
    """Lmfit

    Arguments
    ---------
    sample : sequence
        amplitudes between 0 and 1.

    Keywords
    --------
    fit_model : string
        name of the function (in pdf module) to use for the fit
    bins : string
        method to compute the bin width (inherited from numpy.histogram)
    p0 : dict
        Initial parameters. If None, estimated automatically.
    xtol : float
        ??
    ftol : float
        ??

    Return
    ------
    A Statfit Class
    """
    start = time.time()
    #winsize = len(sample)
    bad = False

    #--------------------------------------------------------------------------
    # Clean sample
    #--------------------------------------------------------------------------
    sample = np.array(sample)
    sample = sample[np.isfinite(sample)]
    if len(sample) == 0:
        bad = True
        sample = np.zeros(10)+1

    #--------------------------------------------------------------------------
    # Make the histogram
    #--------------------------------------------------------------------------
#    n, edges, patches = hist(sample, bins=bins, normed=True)
    n, edges = np.histogram(sample, bins=bins, density=True)
#    plt.clf()

    x = ((np.roll(edges, -1) + edges)/2.)[0:-1]

    #--------------------------------------------------------------------------
    # Initial Parameters for the fit
    #--------------------------------------------------------------------------
    if p0 is None:
        p0 = param0(sample)

    prm0 = Parameters()
    #     (Name,    Value,                 Vary,   Min,    Max,    Expr)
    prm0.add('a',     p0['a'],              True,   0,  1,      None)
    prm0.add('s',     p0['s'],              True,   0,  1,      None)
    prm0.add('mu',    p0['mu'],             True,   .5, 10,    None)
    prm0.add('pt',    np.average(sample)**2,False,  0,  1,      'a**2+2*s**2*mu')

    #if fit_model == 'hk':
    #    # From [Dutt and Greenleaf. 1994, eq.14]
    #    prm0.add('a4',    np.average(sample)**4,False,  0,  1,
    #             '8*(1+1/mu)*s**4 + 8*s**2*s**2 + a**4')

    #--------------------------------------------------------------------------
    # Fit
    #--------------------------------------------------------------------------
    pdf2use = getattr(pdf, fit_model)

    # use 'lbfgs' fit if error with 'leastsq' fit
    try:
        p = minimize(pdf2use, prm0, args=(x, n), method='leastsq',
            xtol=xtol, ftol=ftol)
    except KeyboardInterrupt:
        raise
    except:
        # TODO: do we expect a specific exception?
        print('!! Error with LEASTSQ fit, use L-BFGS-B instead')
        p = minimize(pdf2use, prm0, args=(x, n), method='lbfgs')

    #--------------------------------------------------------------------------
    # Output
    #--------------------------------------------------------------------------
    elapsed = time.time() - start

    values = {}

    # Create values dict For lmfit >0.9.0 compatibility since it is no longer
    # in the minimize output
    for i in p.params.keys():
        values[i] = p.params[i].value

    # Results
    result = Statfit(sample, pdf2use, values, p.params,
             p.chisqr, p.redchi, elapsed, p.nfev, p.message, p.success,
             p.residual, x, n, edges, bins=bins)

    # Identify bad results
    # TODO: consider making this code part of Statfit.py?
    if bad:
        result.success = False
        result.values['a'] = 0
        result.values['s'] = 0
        result.values['mu'] = 0
        result.values['pt'] = 0
        result.chisqr = 0
        result.redchi = 0
        result.message = 'No valid data in the sample'
        result.residual = 0

    return result



"""Various python classes for rsr package
"""

from typing import Callable, Dict
from dataclasses import dataclass

import numpy as np
import matplotlib.pyplot as plt

import subradar as sr
from . import invert
import lmfit

@dataclass
class Statfit:
    """Class holding statistical fit results
    """

    sample: np.ndarray
    # PDF function used
    func: Callable
    # lmfit minimization results
    values: Dict[str, float]
    params: lmfit.Parameters
    chisqr: float
    redchi: float
    elapsed: float
    nfev: float
    message: str
    success: bool
    residual: np.ndarray
    # Histogram bin centers and values
    x: np.ndarray
    n: np.ndarray
    edges: np.ndarray
    bins: np.ndarray

    def power(self, db=True):
        """Total (pt), coherent (pc), and incoherent (pn) components in power
        """
        pt, pc, pn = np.average(self.sample)**2, self.values['a']**2, \
                     2*self.values['s']**2*self.values['mu']
        mu = self.values['mu']
        if db:
            pt, pc, pn = 10*np.log10(pt), 10*np.log10(pc), 10*np.log10(pn)
        if not self.success:
            pt, pc, pn, mu = 0, 0, 0, 0 # TODO: should these be floats?
        return {'pt':pt, 'pc':pc, 'pn':pn, 'pc-pn':pc-pn, 'mu':mu}


    def crl(self, **kwargs):
        """Correlation coefficient between distribution and theoretical fit
        """
        try:
            out = np.corrcoef(self.n, self.n+self.residual)[0,1]
        except: # TODO: make this more specific
            out = 0.

        if (not np.isfinite(out)) or (not self.success):
            out = 0.
        return out


    def invert(self, frq=60e6, th_max=1e-4, cl_logrange=[5], n=1e4,
        ep_range=[1,100], method='iem', approx='Small_S', ):
        """Invert signal components into physical properties
        """
        if method == 'spm':
            out =  getattr(invert, method)(frq, self.power()['pc'],
                   self.power()['pn'])
        if method == 'iem':
            out = sr.invert.power2srf_norminc(method, approx,
            self.power()['pc'], self.power()['pn'], ep_range=ep_range,
            th_max=th_max, wf=frq, cl_logrange=cl_logrange, n=n)
        return out


    def plot(self, ylabel='Normalized Probability', xlabel='Amplitude',
        color='k', ls='-', bins=None, fbins=100, alpha=.1,
        method='compound', histtype='stepfilled', xlim=None):
        """Plot histogram and pdf
        """
        if bins is None:
            bins = self.bins

        # TODO: fix unused histogram arguments fbins, alpha, histtype
        _, edges, _ = plt.hist(self.sample, bins=bins, density=True)
        x = [ val-(val-edges[i-1])/2. for i, val in enumerate(edges) ][1:]

        plt.plot(x, self.func(self.values, x, method=method), color=color,
                 ls=ls, linewidth=2)
        plt.xlim(xlim)
        plt.ylabel(ylabel, size=17)
        plt.xlabel(xlabel, size=17)
        plt.yticks(size='17')
        plt.xticks(size='17')


    def report(self, frq=60e6, inv='spm', file=None):
        """Print a report for the fit
        """
        buff = []
        add = buff.append

        add('[%i] [%6.2f s.] [%3.0f eval.] [%3.3f]\n'
            % (self.flag(), self.elapsed, self.nfev, self.crl()))

        add('%s\n' % (self.message))

        pwr = self.power()
        for key, val in pwr.items():
            add('%s = %3.1f dB, ' % (key, val))

        add('\n')

        p = self.invert(frq=frq, method=inv)
        add('%s @ %.0f MHz gives, eps = %.3f, sh = %.2e m' % (inv.upper(),
            frq*1e-6, p['eps'], p['sh']))

        add('\n')

        out = "".join(buff)
        print(out, file=file)


    def flag(self):
        """0 is bad data, 1 is good data
        """
        out = self.success * np.isfinite(self.crl()) * (self.crl() > 0)
        return int(out)




import numpy as np
import scipy
from . import pdf as  funcs
from . import run


def inverse_cdf(f, xN=1e3, xlim=(0,1), **kwargs):
    """Provides the inverse cumulative density function (CDF) of a probability
    density function

    ARGUMENTS
    ---------
    f : function
        The PDF to evaluate
    xN : integer
        The x-axis sampling of the PDF
    xlim : (integer, integer)
        The x-axis range overwhich the pdf is considered

    RETURN
    ------
    the inverse CDF
    """
    # Probability Density Function
    x = np.linspace(xlim[0],xlim[1],int(xN))
    #func = f(x)
    # Cumulative Distribution Function
    cdf = np.cumsum(f(x))
    # Normalisation to 1
    cdf = cdf/cdf.max()
    # Inverse CDF
    return scipy.interpolate.interp1d(cdf,x)


def sample(f, N=1e3, xN=1e3, xlim=(0,1), precision=None, **kwargs):
    """Provides a random trial X over a given PDF following the Inverse
    Transform Sampling technique so that X=F^{-1}(U), where F^{-1} is the
    inverse cumulative distribution function of the PDF and U is a random
    variable uniformly distributed between 0 and 1.

    ARGUMENTS
    ---------
    f : function
        The PDF to evaluate
    N : Integer
        The number of random evaluation to make
    xN : integer
        The x-axis sampling of the PDF
    xlim : (integer, integer)
        The x-axis range overwhich the pdf is considered
    precision : float
        precision on the measurement [dB]

    RETURN
    ------
    N random evaluation over the considered PDF
    """
    uniform_samples = np.random.random(int(N))
    required_samples = inverse_cdf(f, xN=xN, xlim=xlim)(uniform_samples)

    if precision:
        for i, rsample in enumerate(required_samples):
            linear_precision = np.abs(10**(precision/2/20) -
                                      10**(-precision/2/20))*rsample
            required_samples[i] = np.random.normal(rsample, linear_precision)

    return required_samples


def pdf(func, params, method=None, **kwargs):
    """Provides a random trial X over a PDF in rsr.pdf

    ARGUMENTS
    ---------
    func : string
        PDF name in rsr.pdf
    params : dict
        parameters to be passed to the PDF function
    kwargs : dict
        Any arguments used bu sample.sample

    RETURN
    ------
    N random evaluation over the considered PDF
    """
    if method is None:
        f = lambda x: getattr(funcs, func)(params, x,)
    else:
        f = lambda x: getattr(funcs, func)(params, x, method=method)
    return sample(f, **kwargs)


def rsr(func, params, method=None, **kwargs):
    """Apply the RSR over a set of generated random amplitudes

    ARGUMENTS
    ---------
    func : string
        PDF name in rsr.pdf
    params : dict
        parameters to be passed to the PDF function
    kwargs : dict
        Any arguments used bu sample.sample

    RETURN
    ------
    N random evaluation over the considered PDF
    """
    amp = pdf(func, params, method=method, **kwargs)
    p = run.processor(amp)
    return p


def power_to_params(params):
    """Converts Pc and Pn powers in dB to a params dictionary
    """
    pc_db, pn_db, mu = params['pc'], params['pn'], params['mu']
    pc, pn = 10**(pc_db/10), 10**(pn_db/10)
    params['a'] = np.sqrt(pc)
    params['s'] = np.sqrt(pn/2./mu)
    return params


def params_to_power(params, dB=True):
    """Converts a params dict into pc and pn powers in dB
    """
    mu = params['mu']
    pc = params['a']**2
    pn = 2*params['s']**2*mu
    if dB:
        pc = 10*np.log10(pc)
        pn = 10*np.log10(pn)

    params['pc'] = pc
    params['pn'] = pn
    return params


def effective_precision(func, params, Nsets=1, **kwargs):
    """Gets the effective precision applicable to the derivation of Pc and Pn
    from the rsr algorithm. The precision is the median of Nsets of histograms
    randomly obtained from params.

    ARGUMENTS
    ---------
    func : string
        PDF name in rsr.pdf
    params : dict
        parameters to be passed to the PDF function
    Nsets: integer
        Number of amplitude histogram to draw
    kwargs : dict
        Any arguments used bu sample.sample

    RETURN
    ------
    pc and pn median precisions

   """
    power = params_to_power(params, dB=True)
    ps = [rsr(func, params, **kwargs) for i in np.arange(Nsets)]
    d_pcs = [p.power()['pc']-power['pc'] for p in ps]
    d_pns = [p.power()['pn']-power['pn'] for p in ps]
    out = {
        'ps':ps, 'power':power, 'd_pcs':d_pcs, 'd_pns':d_pns,
        'd_pc':np.median(d_pcs), 'd_pn':np.median(d_pns)
    }
    return out



"""
Wrappers for running RSR processing
"""

import time
import multiprocessing

import numpy as np
import pandas as pd
from sklearn.neighbors import KDTree

from . import fit


def timing(func):
    """Outputs the time a function takes to execute.
    """
    def func_wrapper(*args, **kwargs):
        t1 = time.time()
        func(*args, **kwargs)
        t2 = time.time()
        print("- Processed in %.1f s.\n" % (t2-t1))
    return func_wrapper


def scale(amp):
    """Provide a factor to scale a set of amplitudes between 0 and 1
    for correct rsr processing through fit.lmfit
    """
    y, x = np.histogram(np.abs(amp), bins='fd')
    pik = x[y.argmax()]
    out = 1/(pik*10)
    return out

def processor_mp_(kwargs):
    """ Wrapper function for calling processor with multiprocessing """
    return processor(**kwargs)

def processor(amp, gain=0., bins='stone', fit_model='hk', scaling=True, **kwargs):
    """Apply RSR over a sample of amplitudes

    Arguments
    ---------
    amp: float array
        Linear amplitudes

    Keywords
    --------
    gain: float
        Gain (in dB power) to add to the amplitudes
    bins: string
        Method to compute the bin width (inherited from numpy.histogram)
    fit_model: string
        Name of the function (in pdf module) to use for the fit
    scaling: boolean
        Whether to scale the amplitudes before processing.
        That ensures a correct fit in case the amplitudes are << 1
    output: string
        Format of the output
    Return
    ------
    A Statfit class
    """
    # Remove zero values
    amp = amp[amp > 0]

    # Gain and Scaling
    amp = amp * 10**(gain/20.)
    scale_amp = scale(amp) if scaling else 1
    amp = amp*scale_amp

    # Fit
    a = fit.lmfit( np.abs(amp), bins=bins, fit_model=fit_model)

    # Remove Scaling
    a.sample = amp/scale_amp

    #a0 = a.values['a']/scale_amp
    #b0 = np.sqrt(10**(a.power()['pn']/10))/scale_amp/np.sqrt(2*a.values['mu'])

    a.values['a'] = a.values['a']/scale_amp
    a.values['s'] = a.values['s']/scale_amp

    # Output ID in values if provided
    a.values['ID'] = kwargs.get('ID', -1)

    return a


def cb_processor(a):
    """
    Callback function for processor

    Argument:
    ---------
    a: class
        Results from "processor" (Statfit class)
    """
    p = a.power()
    #print(p)
    print("#%d\tCorrelation: %.3f\tPt: %.1f dB   Pc: %.1f dB   Pn: %.1f dB   mu: %.1f" %
            (a.values['ID'], a.crl(), p['pt'], p['pc'], p['pn'], a.values['mu'] ) )
    return a


def frames(x ,winsize:int=1000, sampling:int=250, **kwargs):
    """
    Defines along-track frames coordinates for rsr application

    Arguments
    ---------
    x: float array
        vector index

    Keywords
    --------
    winsize: int
        Number of elements in a window
    sampling: int
        Sampling step
    """
    # Window first and last id
    xa = x[:int(x.size-winsize):int(sampling)]
    xb = xa + winsize-1

    # Cut last window in limb
    if xb[-1] > x[-1]:
        xb[-1] = x[-1]
    xo = [val+(xb[i]-val)/2. for i, val in enumerate(xa)]

    # Output
    out = [ np.array([xa[i], xb[i]]).astype('int64') for i in np.arange(xa.size)  ]

    out = {'xa':np.array(xa, dtype=np.int64),
           'xb':np.array(xb, dtype=np.int64),
           'xo':np.array(xo, dtype=np.float64),
           }

    return out

#@timing
def along(amp, nbcores=1, verbose=True, **kwargs):
    """
    RSR applied on windows sliding along a vector of amplitudes

    Arguments
    ---------
    amp: Float array
        A vector of amplitudes

    Keywords
    --------
    nbcores: int
        number of cores
    verbose: boolean
        print results
    Any keywords accepted by 'processor' and 'frames'

    Return
    ------

    """
    t1 = time.time()

    #-----------
    # Parameters
    #-----------

    # Windows along-track
    x = np.arange( len(amp) ) #vector index
    w = frames(x, **kwargs)

    # Jobs Definition
    jobs = []
    for ai, bi, xo in zip(w['xa'], w['xb'], w['xo']):
        jobs.append({'amp': amp[ai:bi], **kwargs, 'ID': xo})

    #-----------
    # Processing
    #-----------

    results = []
    async_cb = cb_processor if verbose else None
    if verbose:
        print("#", len(jobs), "jobs")

    if nbcores <= 1:
        # Do NOT use the multiprocessing package
        for job in jobs:
            a = processor(**job)
            if async_cb:
                async_cb(a)
            b = {**a.values, **a.power(), 'crl':a.crl(), 'chisqr':a.chisqr,}
            results.append(b)
    else:
        # Do use the multiprocessing package
        with multiprocessing.Pool(nbcores) as pool:
            for a in pool.imap(processor_mp_, jobs):
                if async_cb:
                    async_cb(a)
                b = {**a.values, **a.power(), 'crl':a.crl(), 'chisqr':a.chisqr,}
                results.append(b)

    out = pd.DataFrame(results)
    out['xa'] = w['xa']
    out['xb'] = w['xb']
    out['xo'] = w['xo']

    if verbose:
        t2 = time.time()
        print("- Processed in %.1f s.\n" % (t2-t1))

    return out


#@timing
def incircles(amp, amp_x, amp_y, circle_x, circle_y, circle_r, leaf_size=None,
              nbcores=1, verbose=True, **kwargs):
    """
    RSR applied over data within circles

    Arguments
    ---------
    amp: Float array
        A vector of amplitudes
    amp_x: Float array
        X coordinates for amp
    amp_y: Float array
        Y coordinates for amp
    circle_x: Float array
        X coordinates for circles
    circle_y: Float array
        Y_coordinates for circles
    circle_r: Float
        Radius of the circles
    leaf_size: Integer (Default: None)
        Set the leaf size for the KD-Tree. Inherits from sklearn.
        If None, use a brute force technique

    Keywords
    --------
    leaf_size: Integer (Default: None)
        Set the leaf size for the KD-Tree. Inherits from sklearn.
        If None, use a brute force technique
    nbcores: int
        number of cores
    verbose: boolean
        print results
    Any keywords accepted by 'processor' and 'frames'

    Return
    ------

    """
    t1 = time.time()

    #-----------
    # Parameters
    #-----------

    # Coordinates units
    #if deg is True:
    #    metrics = 'haversine'
    #    amp_x = np.deg2rad(amp_x)
    #    amp_y = np.deg2rad(amp_y)
    #    circle_x = np.deg2rad(circle_x)
    #    circle_y = np.deg2rad(circle_y)
    #    circle_r = np.deg2rad(circle_r)
    #else:
    #    metrics = 'euclidian'

    # KD-Tree
    if leaf_size is None:
        leaf_size = len(amp)
    amp_xy = np.array(list(zip(amp_x, amp_y)))
    tree = KDTree(amp_xy, leaf_size=leaf_size)

    # Radius Query
    circle_xy = np.array(list(zip(circle_x, circle_y)))
    ind = tree.query_radius(circle_xy, r=circle_r)

    # Jobs Definition
    jobs = []
    for i, data_index in enumerate(ind, start=1):
        if data_index.size == 0:
            continue
        data = np.take(amp, data_index)
        jobs.append({'amp': data, **kwargs, 'ID': i})

    #-----------
    # Processing
    #-----------

    results = []
    async_cb = cb_processor if verbose else None
    if verbose:
        print("#", len(jobs), "jobs")

    if nbcores <= 1:
        # Do NOT use the multiprocessing package
        for job in jobs:
            a = processor(**job)
            if async_cb:
                async_cb(a)
            b = {**a.values, **a.power(), 'crl':a.crl(), 'chisqr':a.chisqr,}
            results.append(b)
    else:
        # Do use the multiprocessing package
        with multiprocessing.Pool(nbcores) as pool:
            for a in pool.imap(processor_mp_, jobs):
                if async_cb:
                    async_cb(a)
                b = {**a.values, **a.power(), 'crl':a.crl(), 'chisqr':a.chisqr,}
                results.append(b)

    out = pd.DataFrame(results)
    #out['xa'] = w['xa']
    #out['xb'] = w['xb']
    #out['xo'] = w['xo']

    if verbose:
        t2 = time.time()
        print("- Processed in %.1f s.\n" % (t2-t1))

    return out




"""Tools and utilities for running the Radar Statistical Reconnaissance
"""

__version__ = "1.0.7"
__author__ = "Cyril Grima"

__all__ = ["detect", "fit", "invert", "pdf", "run", "utils"]

from . import detect, fit, invert, pdf, run, sample, utils



"""
Various tools for application of the rsr
Author: Cyril Grima <cyril.grima@gmail.com>
"""

import time

import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt

#from . import pdf
from . import fit
#from . import invert


def inline_estim(vec, fit_model='hk', bins='auto', inv='spm', winsize=1000.,
                 sampling=250., frq=60e6, save='.inline_estim_last',
                 verbose=True, **kwargs):
    """Histogram statistical estimation over windows sliding along a vector

    Arguments
    ---------
    vec : sequence
        A vector of linear amplitude values

    Keywords
    --------
    fit_model : string
        pdf to use to estimate the histogram statistics (in fit.)
    bins : string
        method to compute the bin width (inherited from astroML.plotting.hist)
    inv : string
        inversion method (in invert.)
    winsize : int
        number of amplitude values within a window
    sampling : int
        window repeat rate
    frq : float
        Radar center frequency [Hz]
    save : string
        file name (without extension) to save the results in an ascii file
    verbose : bool
        Display fit results informations
    """
    start = time.time()

    #--------------------------------------------------------------------------
    # Windows along-track
    #--------------------------------------------------------------------------
    x = np.arange(vec.size) #vector index
    #xa = x[:x.size-winsize:sampling] #windows starting coordinate
    xa = x[:np.int(x.size-winsize):np.int(sampling)] #windows starting coordinate
    xb = xa + winsize-1 #window end coordinate
    if xb[-1] > x[-1]:
        xb[-1] = x[-1] #cut last window in limb
    xo = [val+(xb[i]-val)/2. for i, val in enumerate(xa)]

    #--------------------------------------------------------------------------
    # Table to hold the results
    #--------------------------------------------------------------------------
    columns = ['xa', 'xb', 'xo', 'pt', 'pc', 'pn', 'crl', 'chisqr', 'mu',
               'eps', 'sh', 'flag']
    index = np.arange(xa.size)
    table = DataFrame({'xa':xa, 'xb':xb, 'xo':xo},
                      index=index, columns=columns)

    #--------------------------------------------------------------------------
    # Fit for all the windows
    #--------------------------------------------------------------------------
    for i, val in enumerate(xo):
        if verbose:
            print('ITER '+ str(i+1) + '/' + str(xa.size) +
            ' (observations ' + str(xa[i]) + ':' + str(xb[i]) + ')')

        sample = vec[int(xa[i]):int(xb[i])]
        p = fit.lmfit(sample, fit_model=fit_model, bins=bins)

        table.loc[i, 'pt'] = p.power()['pt']
        table.loc[i, 'pc'] = p.power()['pc']
        table.loc[i, 'pn'] = p.power()['pn']
        table.loc[i, 'crl'] = p.crl()
        table.loc[i, 'chisqr'] = p.chisqr
        table.loc[i, 'mu'] = p.values['mu']
        table.loc[i, 'eps'] = p.invert(frq=frq, method=inv)['eps']
        table.loc[i, 'sh'] = p.invert(frq=frq, method=inv)['sh']
        table.loc[i, 'flag'] = p.flag()

        if verbose:
            p.report(frq=frq)

    #--------------------------------------------------------------------------
    # Output
    #--------------------------------------------------------------------------
    elapsed = time.time() - start
    print('DURATION: %4.1f min.' % (elapsed/60.))

    if save is not None:
        ext = '.'+fit_model+'.'+inv
        table.to_csv(save+ext+'.txt', sep='\t', index=False, float_format='%.3f')

    # Convert x coordinates to integers
    table['xa'] = table['xa'].astype(np.int64)
    table['xo'] = table['xo'].astype(np.int64)
    table['xb'] = table['xb'].astype(np.int64)

    return table


def plot_inline(a, frq=60e6, title=''):
    """Plot infos from a DataFrame ceated by inline_estim

    Arguments
    ---------
    a : DataFrame
        inline_estim output
    """
    w = np.where(a.flag)
    x = a.xo.values.astype('float')[w]
    crl = a.crl.values.astype('float')[w]
    pt = a.pt.values.astype('float')[w]
    pc = a.pc.values.astype('float')[w]
    pn = a.pn.values.astype('float')[w]
    eps = a.eps.values.astype('float')[w]
    sh = a.sh.values.astype('float')[w]

    plt.figure(figsize=(15,10))

    #--------------------------------------------------------------------------
    # Correlation Coefficient
    #--------------------------------------------------------------------------
    ax_crl = plt.subplot2grid((5, 1), (0, 0))
    plt.plot(x, crl, 'o-', color='k')
    plt.grid(alpha=.5)
    plt.ylabel(r'Correl. Coeff.', size=17)
    plt.xticks(size='10')
    plt.yticks(size='15')
    plt.title(title, size='15')

    #--------------------------------------------------------------------------
    # Signal components
    #--------------------------------------------------------------------------
    ax_pwr = plt.subplot2grid((5,1), (1, 0), rowspan=2)
    ax_pwr.fill_between(x, pc, pn, where=pc>=pn, facecolor='k', alpha=.05, interpolate=True)
    ax_pwr.fill_between(x, pc, pn, where=pc<=pn, facecolor='k', alpha=.4, interpolate=True)
    plt.plot(x, pc, color='k', lw=3, alpha=.9, label=r'Reflectance $(P_c)$')
    plt.plot(x, pn, color='k', lw=3, alpha=.6, label=r'Scattering $(P_n)$')
    plt.ylim([-40,0])
    plt.grid(alpha=.5)
    plt.ylabel(r'Power $[dB]$', size=17)
    plt.yticks(size='15')
    plt.xticks(size='10')
    plt.legend(loc='lower right', fancybox=True).get_frame().set_alpha(0.5)

    #--------------------------------------------------------------------------
    # Permittivity
    #--------------------------------------------------------------------------
    ax_eps = plt.subplot2grid((5,1), (3, 0), rowspan=2)
    plt.semilogy(x, eps, color='k', lw=3, alpha=.9, label=r'Permittivity $(\epsilon)$')
    plt.ylim(1,100)
    plt.grid(True, which='both', alpha=.5)
    plt.ylabel('Permittivity', size=17)
    plt.xticks(size='10')
    plt.xlabel('Frame #', size=12)
    plt.yticks(size='15')
    ax_eps.set_yticks([1, 10, 100])
    ax_eps.set_yticklabels(['1', '10', '100'])

    #--------------------------------------------------------------------------
    # RMS height
    #--------------------------------------------------------------------------
    ax_sh = ax_eps.twinx()
    plt.semilogy(x, sh, '-', color='k', lw=3, alpha=.3, label=r'RMS height $(\sigma_h)$')
    plt.semilogy(x, eps, color='k', lw=3, alpha=.9, label=r'Permittivity $(\epsilon)$')
    plt.ylim(0.01,1)
    plt.ylabel(r'RMS height $[m]$', size=17)
    plt.yticks(size='15')
    ax_sh.set_yticks([.01, .1, 1])
    ax_sh.set_yticklabels(['0.01', '0.1', '1'])
    ax_sh.set # TODO: what is this
    plt.legend(loc='upper right', fancybox=True).get_frame().set_alpha(0.5)


def plot_along(a, title=''):
    """Plot infos from a DataFrame created by run.along

    Arguments
    ---------
    a : Pandas DataFrame
        run.along output
    """
    f, ax = plt.subplots(2, figsize=(16, 16), dpi= 80, )#wspace=0, hspace=0)

    x = a['xo'].values
    pc = a['pc'].values
    pn = a['pn'].values
    crl = a['crl'].values
    chisqr = a['chisqr'].values

    for i in ax[1:2]:
        i.grid()
        i.xaxis.label.set_size(15)
        i.yaxis.label.set_size(15)
        i.tick_params(labelsize=15)
        i.title.set_size(20)

    ax[0].plot(x, pc, 'k', lw=3, label='$P_c$')
    ax[0].plot(x, pn, '.6', lw=3, label='$P_n$')
    ax[0].fill_between(x, pc, pn, where=pc >= pn, alpha=.2, label='Dominantly Specular')
    ax[0].fill_between(x, pc, pn, where=pc <= pn, alpha=.2, label='Dominatly Diffuse')
    ax[0].set_title('RSR-derived Coherent and Incoherent Energies', fontweight="bold", fontsize=20)
    ax[0].set_ylabel('$[dB]$')
    ax[0].set_xlim(0, x.max())
    ax[0].legend(loc=3, ncol=2, fontsize='large')

    ax_chisqr = ax[1].twinx()
    ax_chisqr.plot(x, chisqr, '.6', lw=3)
    ax_chisqr.set_ylabel('Chi-square', color='.6')
    ax_chisqr.yaxis.label.set_size(15)
    ax_chisqr.tick_params(labelsize=15)

    ax[1].plot(x, crl, 'k', lw=3)
    ax[1].set_title('Quality Metrics', fontweight="bold", fontsize=20)
    ax[1].set_ylabel('Correlation Coefficient')
    ax[1].set_xlim(0, x.max())
    ax[1].set_ylim(0, 1.1)
    ax[1].legend(loc=3, ncol=2, fontsize='large')
    ax[1].set_xlabel('Bin #')


def grid_coordinates(r, xlim, ylim, shape='square'):
    """Provides point coordinates agenced within a pattern

    Arguments
    ---------
    r : Integer
        Spacing between two consecutive points

    xlim : (float, float)
           Minimum and maximum values of x coordinates

    ylim : (float, float)
           Minimum and maximum values of y coordinates

    shape : String
            Shape of the grid mesh. If 'hexagonal', r define the x spacing

    """
    if shape == 'square':
        dx = r
        dy = r
    if shape == 'hexagonal':
        dx = r
        dy = np.sqrt(3)*dx/2.
    x = np.arange(xlim[0], xlim[1], dx)
    y = np.arange(ylim[0], ylim[1], dy)
    X, Y = np.meshgrid(x, y)
    if shape == 'hexagonal':
        X[::2] = X[::2] + dx/2.

    return X.flatten(), Y.flatten()




"""
Various tools for detecting specific signals in a radargram
"""

import numpy as np

def surface(rdg, method='maxvec', loc=None, ywinwidth=None):
    """Detect surface echo in a radargram

    Arguments
    ---------
    rdg : Array
        Radargram

    Keywords
    --------
    method : string
        Method to use for surface detection
    loc : vec
        A coordinate guess of where the surface should be
    ywinwidth : [int, int]
        Window limit around loc where the algo should search for the surface

    Output
    ------
    y : coordinate for the surface location
    val : value of te surface echo
    ok : Whether the detection is estimated to be good (1) or not (0)
    """

    n = rdg.shape

    loc = np.zeros(n[0]).astype('int') if loc is None else loc.astype('int')
    ywinwidth = np.array([0,n[1]-1]).astype('int') if ywinwidth is None else ywinwidth.astype('int')

    if ywinwidth[0] < 0:
        ywinwidth[0] = 0
    if ywinwidth[-1] > n[1]-1:
        ywinwidth[-1] = n[1]-1

    return globals()[method](rdg, loc, ywinwidth)


def maxvec(rdg, loc, ywinwidth):
    """
    Select the maximum value for each range line in a radargram
    """
    n = len(loc)
    y = np.zeros(n).astype('int')
    val = np.zeros(n).astype('int')
    ok = np.zeros(n).astype('int')
    for i, loci in enumerate(loc):
        pls = np.abs(rdg[i, :])
        itv = pls[loci+ywinwidth[0]:loci+ywinwidth[1]]
        y[i] = loci + ywinwidth[0] + np.argmax(itv)
        val[i] = pls[y[i]]
        # No quality metrics for now
        ok[i] = 1
    return {'y':y, 'val':val, 'ok':ok}


def maxprd(rdg, loc, ywinwidth):
    """
    Select the maximum of the val*val/dt for each range line in a radargram
    """
    n = len(loc)
    y = np.zeros(n).astype('int')
    val = np.zeros(n).astype('int')
    ok = np.zeros(n).astype('int')
    for i, loci in enumerate(loc):
        pls = np.abs(rdg[i, :])
        prd = np.abs(np.roll(np.gradient(pls), 2) * pls)
        itv = prd[loci+ywinwidth[0]:loci+ywinwidth[1]]
        y[i] = loci + ywinwidth[0] + np.argmax(itv)
        val[i] = pls[y[i]]
        # No Quality metrics for now
        ok[i] = 1
    return {'y':y, 'val':val, 'ok':ok}
